
/*
 * bwx.anal.c:
 * routines to collect dynamic data about shift instructions
 * which can be replaced with SEXTB, SEXTW, LDBU, or LDWU
 * instructions.
 *
 * Copyright (C) 1999	Rick Gorton gorton@world.std.com
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdio.h>
#include "datatypes.h"

/*
 * Data declarations local to this file
 */

/*
 * Initialization
 */
void DynamicApplicationInit()
{
  int	i;

  for (i=0; i<= (int)LAST_OP; i++)
    {
    dyn_count[i] = 0x0;
    }

 dyn_SEXTB.counter = 0x0;
 dyn_SEXTW.counter = 0x0;
 dyn_SEXTL.counter = 0x0;
 dyn_LDBU.counter = 0x0;
 dyn_LDWU.counter = 0x0;
 dyn_STB.counter = 0x0;
 dyn_STW.counter = 0x0;
 dyn_AMASK.counter = 0x0;
 dyn_IMPLVER.counter = 0x0;
 total_dyn_count = 0x0;
}

/*
 * Print out results
 */
void DumpDynamicData()
{
  int   i;

  fprintf(Inst_Logfile, "*** dyn data begin ***\n");
  fprintf(Inst_Logfile, "    Potential BWX candidates\n");
  fprintf(Inst_Logfile, "\tSEXTB candidates:\t%ld\n", dyn_SEXTB.counter);
  fprintf(Inst_Logfile, "\tSEXTW candidates:\t%ld\n", dyn_SEXTW.counter);
  fprintf(Inst_Logfile, "\tSEXTL candidates (ADDL):\t%ld\n", dyn_SEXTL.counter);
  fprintf(Inst_Logfile, "\tLDBU candidates:\t%ld\n", dyn_LDBU.counter);
  fprintf(Inst_Logfile, "\tLDWU candidates:\t%ld\n", dyn_LDWU.counter);
  fprintf(Inst_Logfile, "\tSTB candidates:\t%ld\n", dyn_STB.counter);
  fprintf(Inst_Logfile, "\tSTW candidates:\t%ld\n", dyn_STW.counter);

	// Hmmm.  To report AMASK & IMPLVER or not?
	// If they occur a lot, it means the program wasn't compiled with
	// a specific architecture in mind...
	// For now, leave it in
  fprintf(Inst_Logfile, "\tAMASK uses:\t%ld\n", dyn_AMASK.counter);
  fprintf(Inst_Logfile, "\tIMPLVER uses:\t%ld\n", dyn_IMPLVER.counter);

  fprintf(Inst_Logfile, "\tTotal dynamic instructions executed:\t%ld\n", total_dyn_count);
  fprintf(Inst_Logfile, "*** dyn data end ***\n");
  fprintf(Inst_Logfile, "*** Summary ***\n");

	/*
	 * IF the program ran a while (set a threshold of about a second
	 * actually try to guess if there is a real performance problem.
	 * The longer the program runs, the more important these sequence become.
	 * So, 
	 */

  if (total_dyn_count <= 500.0e6)
    {
    fprintf (Inst_Logfile,
	"BWX instruction usage (or lack thereof) is not a performance problem\n");
    }
  else
    {
    double	threshold;
    long	net_bwx, curr_dyn_count;
    double	ratio;

    threshold = 0.01;	// Start it at 1%
    net_bwx = dyn_SEXTB.counter + dyn_SEXTW.counter + dyn_SEXTL.counter
		+ dyn_LDBU.counter + dyn_LDWU.counter
		+ dyn_STB.counter + dyn_STW.counter;
		// Don't bother counting AMASK/IMPLVER for now
    ratio = (double) net_bwx / (double) total_dyn_count;
    curr_dyn_count = total_dyn_count;
    while (curr_dyn_count > 0x3b9aca00)	// 0x3b9aca00 == 1.0e9
	{
	// Lower the threshold as the execution time goes up
	// Half the effective execution count
	curr_dyn_count = curr_dyn_count >> 1;
	// And lower the threshold by a bit less than half.  Guess .75 for now
	threshold = threshold * 0.75;
	}

	// And finally, scale it back assuming roughly 1.5x issue rate
	// (nobody is perfect), and most dual issue machines don't do that well
    threshold = threshold * 1.5e0;

    fprintf (Inst_Logfile, "dynamic threshold: %f\n", threshold);
    fprintf (Inst_Logfile, "dynamic ratio: %f\n", ratio);

    if (ratio > threshold)
	{
	fprintf (Inst_Logfile,
"There is probably a significant performance gain to be had by rebuilding this program\n"
"using -arch ev56 (Compaq compilers) or -mcpu=21164 (GCC).  A total of %ld excess\n"
"instructions were executed.  At 500Mhz (single issue), they account for a\n"
"minimum of %ld seconds of execution time, some of which operate on memory.\n",
		total_dyn_count, (long)(total_dyn_count/500.0e6));
	}
    else if ((ratio * 10.0e0) > threshold)
	{
	fprintf (Inst_Logfile,
"BWX instruction usage (or lack thereof) is probably not a performance problem\n");
	}
    else fprintf (Inst_Logfile,
"BWX instruction usage (or lack thereof) is not a performance problem\n");
    }
}


/*
 * Block data collection
 */
void Total_Inst_Count (long this_many)
{
  total_dyn_count += this_many;
}

/*
 * Instruction data collection
 */
void SEXTB_Count()
{
  dyn_SEXTB.counter++;
}

void SEXTW_Count()
{
  dyn_SEXTW.counter++;
}

void SEXTL_Count()
{
  dyn_SEXTL.counter++;
}

void LDBU_Count()
{
  dyn_LDBU.counter++;
}

void LDWU_Count()
{
  dyn_LDWU.counter++;
}

void STB_Count()
{
  dyn_STB.counter++;
}

void STW_Count()
{
  dyn_STW.counter++;
}

void AMASK_Count()
{
  dyn_AMASK.counter++;
}

void IMPLVER_Count()
{
  dyn_IMPLVER.counter++;
}

